@@ -586,7 +586,7 @@ ProcType = Callable[['Unit', ByteString], Optional[Union[DataType, Iterable[Data
 
 def UnitProcessorBoilerplate(operation: ProcType[ByteString]) -> ProcType[Chunk]:
     @wraps(operation)
-    def wrapped(self, data: ByteString) -> Optional[Union[Chunk, Iterable[Chunk]]]:
+    def wrapped(self: Unit, data: ByteString) -> Optional[Union[Chunk, Iterable[Chunk]]]:
         ChunkType = Chunk
         if data is None:
             data = B''
@@ -1186,6 +1186,8 @@ class Unit(UnitBase, abstract=True):
                     yield from (x for x in result if x is not None)
                 elif result is not None:
                     yield result
+            except KeyboardInterrupt:
+                raise
             except BaseException as B:
                 result = self._exception_handler(B, data)
                 message = str(B).strip() or 'unknown'
@@ -1478,9 +1480,9 @@ class Unit(UnitBase, abstract=True):
                 return message
             if isbuffer(message):
                 import codecs
-                pmsg: str = codecs.decode(message, cls.codec, errors='backslashreplace')
+                pmsg: str = codecs.decode(message, cls.codec, 'surrogateescape')
                 if not pmsg.isprintable():
-                    pmsg = message.hex()
+                    pmsg = message.hex().upper()
                 return pmsg
             else:
                 import pprint
@@ -1491,7 +1493,7 @@ class Unit(UnitBase, abstract=True):
             from refinery.lib.tools import get_terminal_size
             message = shorten(
                 message,
-                get_terminal_size() - len(cls.name) - 14,
+                get_terminal_size(75) - len(cls.name) - 14,
             )
         return message
 
