@@ -6,64 +6,72 @@ import itertools
 
 from refinery.units import arg, Unit
 
-from refinery.lib.meta import ByteStringWrapper, SizeInt, metavars
-from refinery.lib.structures import EOF, StructReader
+from refinery.lib.meta import SizeInt, metavars
+from refinery.lib.structures import EOF, StructReader, StreamDetour
 from refinery.lib.argformats import ParserError, PythonExpression
 from refinery.lib.types import INF
-from refinery.lib.tools import isbuffer
 
 
 def identity(x):
     return x
 
 
+_SHARP = '#'
+
+
 class struct(Unit):
     """
-    Read structured data from the beginning of a chunk and store the extracted fields in chunk meta variables.
-    The structure format is specified in extended Python struct format, and all remaining arguments to this unit
-    are the names of the variables that receive the values from this struct. The extended struct format supports
-    all field types supported by Python, as well as the following:
+    Read structured data from the beginning of a chunk and store the extracted fields in chunk meta
+    variables. The structure format is specified in extended Python struct format, and all
+    remaining arguments to this unit are the names of the variables that receive the values from
+    this struct. The extended struct format supports all field types supported by Python, as well
+    as the following:
 
     - `a` for null-terminated ASCII strings,
     - `u` for null-terminated UTF16 strings,
-    - `$` only as the last character, to read all remaining data.
 
-    For example, the string `LLxxHaa$` will read two unsigned 32bit integers, then skip two bytes, then read one
-    unsigned 16bit integer, then two null-terminated ASCII strings and finally, all data that remains. The unit
-    defaults to using native byte order with no alignment.
+    For example, the string `LLxxHaa` will read two unsigned 32bit integers, then skip two bytes,
+    then read one unsigned 16bit integer, then two null-terminated ASCII strings. The unit defaults
+    to using native byte order with no alignment.
 
-    The `spec` parameter may additionally contain named fields `{name:format}`. Here, `format` can either be an
-    integer expression specifying a number of bytes to read, or any single format string character. Parsing
-    such a field will make the parsed data available as a meta variable under the given name. For example, the
-    expression `LLxxH{foo:a}{bar:a}$` would parse the same data as the previous example, but the two ASCII
-    strings would also be output as meta variables under the names `foo` and `bar`, respectively.
-
-    The `format` string of a named field is itself parsed as a foramt string expression, where all the previously
-    parsed fields are already available. For example, `L{:{0}}` reads a single 32-bit integer length prefix and
+    The `spec` parameter may additionally contain format expressions of the form `{name:format}`.
+    Here, `format` can either be an integer expression specifying a number of bytes to read, or any
+    format string. If `name` is specified for an extracted field, its value is made available as a
+    meta variable under the given name. For example, the expression `LLxxH{foo:a}{bar:a}` would be
+    parsed in the same way as the previous example, but the two ASCII strings would also be stored
+    in meta variables under the names `foo` and `bar`, respectively. The `format` string of a named
+    field is itself parsed as a foramt string expression, where all the previously parsed fields
+    are already available. For example, `I{:{}}` reads a single 32-bit integer length prefix and
     then reads as many bytes as that prefix specifies.
 
-    The output arguments are refinery-specific binary format strings that control what the unit outputs:
+    A second format string expression is used to specify the output format. For example, the format
+    string `LLxxH{foo:a}{bar:a}` together with the output format `{foo}/{bar}` would parse data as
+    before, but the output body would be the concatnation of the field `foo`, a forward slash, and
+    the field `bar`. Variables used in the output expression are not included as meta variables. As
+    format fields in the output expression, one can also use `{1}`, `{2}` or `{-1}` to access
+    extracted fields by index. The value `{0}` represents the entire chunk of structured data. By
+    default, the output format `{#}` is used, which represents either the last byte string field
+    that was extracted, or the entire chunk of structured data if none of the fields were extracted.
 
-    - `{0}` places the entire processed data at this position in the output.
-    - `{1}` places the first extracted field at this position in the output.
-    - `{-1}` places the last extracted field at this position in the output.
-    - `{F}` places the field named `F` at this position in the output.
-    - `{$}` represents the last named field or dollar symbol that was parsed.
-
-    The format specifications of binary format strings are refinery pipelines. For example, `{F:b64|zl}` will be
-    the base64-decoded and inflate-decompressed contents of the data that was read as field `F`.
+    Finally, reverse `refinery.lib.argformats.multibin` expressions can be used to post-process the
+    fields included in any output format. For example, `{F:b64:zl}` will be the base64-decoded and
+    inflate-decompressed contents of the data that was read as field `F`.
     """
+
     def __init__(
         self,
         spec: arg(type=str, help='Structure format as explained above.'),
         *outputs: arg(metavar='output', type=str, help='Output format as explained above.'),
+        multi: arg.switch('-m', help=(
+            'Read as many pieces of structured data as possible intead of just one.')) = False,
+        count: arg.number('-n', help=(
+            'A limit on the number of chunks to read in multi mode; default is {default}.')) = INF,
         until: arg('-u', metavar='E', type=str, help=(
-            'An expression evaluated on each chunk. Continue parsing only if the result is nonzero.')) = None,
-        count: arg.number('-c', help=(
-            'A limit on the number of chunks to read. The default is {default}.')) = INF,
+            'An expression evaluated on each chunk in multi mode. New chunks will be parsed '
+            'only if the result is nonzero.')) = None,
     ):
-        outputs = outputs or ['{$}']
-        super().__init__(spec=spec, outputs=outputs, until=until, count=count)
+        outputs = outputs or [F'{{{_SHARP}}}']
+        super().__init__(spec=spec, outputs=outputs, until=until, count=count, multi=multi)
 
     def process(self, data: bytearray):
         formatter = string.Formatter()
@@ -82,13 +90,8 @@ class struct(Unit):
                 spec = byteorder + spec
             return spec
 
-        mainspec, dollar, _empty = mainspec.partition('$')
-        if _empty:
-            raise ValueError('The format string {mainspec}${end} is invalid, a dollar symbol must be the last character.')
-        if dollar:
-            mainspec += '{}'
-
-        for index in itertools.count():
+        it = itertools.count() if self.args.multi else (0,)
+        for index in it:
 
             if reader.eof:
                 break
@@ -115,9 +118,9 @@ class struct(Unit):
                         except ParserError:
                             pass
                     if spec == '':
-                        value = reader.read()
+                        last = value = reader.read()
                     elif isinstance(spec, int):
-                        value = reader.read_bytes(spec)
+                        last = value = reader.read_bytes(spec)
                     else:
                         value = reader.read_struct(fixorder(spec))
                         if not value:
@@ -135,12 +138,12 @@ class struct(Unit):
                         value = value.decode('latin1')
                     if conversion == 't':
                         value = datetime.datetime.utcfromtimestamp(value).isoformat(' ', 'seconds')
-                    if isbuffer(value):
-                        last = ByteStringWrapper(value)
-                    elif isinstance(value, ByteStringWrapper):
-                        last = value
+
                     args.append(value)
-                    if name.isdecimal():
+
+                    if name == _SHARP:
+                        raise ValueError('Extracting a field with name # is forbidden.')
+                    elif name.isdecimal():
                         index = int(name)
                         limit = len(args) - 1
                         if index > limit:
@@ -155,22 +158,15 @@ class struct(Unit):
                     self.log_info(F'the expression ({until}) evaluated to zero; aborting.')
                     break
 
-                size = reader.tell() - checkpoint
-                reader.seek(checkpoint)
+                with StreamDetour(reader, checkpoint) as detour:
+                    full = reader.read(detour.cursor - checkpoint)
+                if last is None:
+                    last = full
 
                 for template in self.args.outputs:
-                    full = reader.read(size)
-                    if last:
-                        last = last.binary
-                    else:
-                        for k in range(len(args) - 1, -1, -1):
-                            if isbuffer(args[k]):
-                                last = args[k]
-                                break
-                        else:
-                            last = B''
-                    output = meta.format_bin(template, self.codec, [full, *args], {'$': last})
-                    for _, key, _, _ in formatter.parse(template):
+                    used = set()
+                    output = meta.format(template, self.codec, [full, *args], {_SHARP: last}, True, used=used)
+                    for key in used:
                         meta.pop(key, None)
                     yield self.labelled(output, **meta)
 
